// Copyright (C) 2012 Denis Shienkov <denis.shienkov@gmail.com>
// Copyright (C) 2012 Laszlo Papp <lpapp@kde.org>
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
// Qt-Security score:significant reason:default

#ifndef QSERIALPORTGLOBAL_H
#define QSERIALPORTGLOBAL_H

#include <QtCore/qstring.h>
#include <QtCore/qglobal.h>
#include <QtSerialPort/qtserialportexports.h>

#endif // QSERIALPORTGLOBAL_H
